"""general configuration
"""

from enum import Enum, auto
from typing import NamedTuple, Optional
from os import getenv

YEAR_START = 2015
MONTH_START = 12
DAY_START = 9


class ParamsVarEnv(NamedTuple):
    key: str

    @property
    def value(self) -> Optional[str]:
        return getenv(self.key)


class VarEnv(ParamsVarEnv, Enum):
    TYPE_STORE = ParamsVarEnv(key="NYSMIX_STORE")
    FOLDER_ZIP = ParamsVarEnv(key="NYSMIX_FOLDER_ZIP")
    FOLDER_CSV = ParamsVarEnv(key="NYSMIX_FOLDER_CSV")
    TYPE_DB = ParamsVarEnv(key="NYSMIX_DB")


"""
defining pipelines
"""

from dataclasses import dataclass
from nysmix.store.base import Store
from nysmix.base import Day, DAY_START, from_name_file_csv
from typing import List
from logging import info
from pathlib import Path

@dataclass
class Extract:
    store_zip: Store
    store_csv: Store

    @property
    def days_next(self) -> List[Day]:
        info("finding next csvs to extract")
        days_next: List[Day] = []
        names_files_zip = [
            name_file
            for name_file in self.store_zip.names_files
            if Path(name_file).suffix == ".zip"
        ]
        names_files_csv = [
            name_file
            for name_file in self.store_csv.names_files
            if Path(name_file).suffix == ".csv"
        ]

        if len(names_files_zip) == 0:
            return []

        # add the latest day
        day_latest = (
            from_name_file_csv(max(names_files_csv))
            if len(names_files_csv) > 0
            else DAY_START
        )
        days_next.append(day_latest)
        # add all later days in the same month
        yearmonth_latest = day_latest.yearmonth
        yearmonth_latest.set_store(self.store_zip)
        days_next.extend(
            [
                from_name_file_csv(name_file_csv)
                for name_file_csv in yearmonth_latest.names_files_archived
                if name_file_csv > day_latest.name_file_csv
            ]
        )
        yearmonth = yearmonth_latest.next
        # add all days from later months
        while yearmonth.name_file_zip in names_files_zip:
            yearmonth.set_store(self.store_zip)
            days_next.extend(
                [
                    from_name_file_csv(name_file_csv)
                    for name_file_csv in yearmonth.names_files_archived
                ]
            )
            yearmonth = yearmonth.next
        info(f"next days to pull: {days_next}")
        return days_next

    def add_next(self) -> None:
        for day in self.days_next:
            day.set_store(self.store_csv)
            day.store_extract_from_zip(store_zip=self.store_zip)


"""
defining pipelines
"""

from dataclasses import dataclass
from nysmix.store.base import Store
from nysmix.base import YearMonth, YEARMONTH_START
from logging import info

@dataclass
class Pull:
    """
    keep the data files (zips) up to date
    """

    store: Store

    @property
    def yearmonth_next(self) -> YearMonth:
        info("finding next month to pull")
        names_files = self.store.names_files
        yearmonth_next = YEARMONTH_START
        while (
            yearmonth_next.name_file_zip in names_files and not yearmonth_next.next.is_future
        ):
            yearmonth_next = yearmonth_next.next
        info(f"next month to pull: {yearmonth_next}")
        return yearmonth_next

    def add_next(self) -> None:
        yearmonth_next = self.yearmonth_next
        yearmonth_next.set_store(self.store)
        yearmonth_next.store_zip_from_site()



"""base classes for nyiso fuel mix data
"""


from dataclasses import dataclass
from datetime import datetime, date, timedelta, timezone
from logging import info
from requests import get
from .store.base import Store
from typing import Optional, Set
from .config.base import YEAR_START, MONTH_START, DAY_START as DAY_START_config
from zipfile import ZipFile
from io import BytesIO

FORMAT_DATE = "%Y%m%d"
TZ = timezone(timedelta(hours=-5))


@dataclass
class Object:
    year: int
    month: int
    day: int = 1

    def __post_init__(self) -> None:
        self.store: Optional[Store] = None

    def set_store(self, store: Store) -> None:
        self.store = store

    @property
    def has_store(self) -> bool:
        has_store = self.store is not None
        if not has_store:
            info("no store set")
        else:
            info(f"store set: {self.store}")
        return has_store

    @property
    def date(self) -> date:
        return date(year=self.year, month=self.month, day=self.day)

    @property
    def str_date(self) -> str:
        return self.date.strftime(FORMAT_DATE)


@dataclass
class YearMonth(Object):
    @property
    def next(self):  # -> YearMonth
        if self.month == 12:
            return YearMonth(
                year=self.year + 1,
                month=1,
            )
        else:
            return YearMonth(
                year=self.year,
                month=self.month + 1,
            )

    @property
    def is_future(self) -> bool:
        return self.year > datetime.now(TZ).year or (
            self.year == datetime.now(TZ).year and self.month > datetime.now(TZ).month
        )

    @property
    def name_file_zip(self) -> str:
        return f"{self.str_date}rtfuelmix_csv.zip"

    @property
    def url(self) -> str:
        return f"http://mis.nyiso.com/public/csv/rtfuelmix/{self.name_file_zip}"

    @property
    def bytes_zip_from_site(self) -> bytes:
        info(f"attempt to download {self.url}")
        with get(self.url) as r:
            content = r.content
        return content

    def store_zip_from_site(self) -> None:
        if self.has_store:
            assert self.store is not None
            self.store.put(content=self.bytes_zip_from_site, name=self.name_file_zip)

    @property
    def bytes_zip_from_store(self) -> bytes:
        if not self.has_store:
            raise Exception(f"no store set for {self}")
        assert self.store is not None
        return self.store.fetch(name=self.name_file_zip)

    @property
    def names_files_archived(self) -> Set[str]:
        with BytesIO(self.bytes_zip_from_store) as f:
            with ZipFile(f) as zf:
                names_files_archived = {info.filename for info in zf.filelist}
        return names_files_archived


YEARMONTH_START = YearMonth(YEAR_START, MONTH_START)


@dataclass(eq=True, order=True)
class Day(Object):
    @property
    def yearmonth(self) -> YearMonth:
        return YearMonth(year=self.year, month=self.month)

    @property
    def name_file_csv(self) -> str:
        return f"{self.str_date}rtfuelmix.csv"

    def bytes_csv_from_zip(self, store_zip: Store) -> bytes:
        yearmonth = self.yearmonth
        yearmonth.set_store(store_zip)
        with BytesIO(yearmonth.bytes_zip_from_store) as f:
            with ZipFile(f) as zf:
                content = zf.read(self.name_file_csv)
        return content

    def store_extract_from_zip(self, store_zip: Store) -> None:
        if self.has_store:
            assert self.store is not None
            self.store.put(content=self.bytes_csv_from_zip(store_zip=store_zip), name=self.name_file_csv)

    @property
    def bytes_csv_from_store(self) -> bytes:
        if not self.has_store:
            raise Exception(f"no store set for {self}")
        assert self.store is not None
        return self.store.fetch(name=self.name_file_csv)


DAY_START = Day(year=YEAR_START, month=MONTH_START, day=DAY_START_config)


def from_name_file_csv(name_file_csv: str) -> Day:
    date_name_file = datetime.strptime(name_file_csv[:8], FORMAT_DATE)
    return Day(
        year=date_name_file.year, month=date_name_file.month, day=date_name_file.day
    )



"""
helper functions for defining stores
"""

from .base import Store, memory, Local
from .google import GoogleCloudStorage
from enum import Enum, auto
from typing import Optional
from pathlib import Path
from nysmix.config.base import VarEnv
from logging import info


class TypeStore(Enum):
    MEMORY = auto()
    LOCAL = auto()
    GOOGLE = auto()


def get_store(type_store: TypeStore, folder: Optional[Path] = None) -> Optional[Store]:
    store: Optional[Store] = None
    if type_store == TypeStore.MEMORY:
        store = memory
    elif type_store == TypeStore.LOCAL:
        assert folder is not None
        store = Local(folder=folder)
    elif type_store == TypeStore.GOOGLE:
        assert folder is not None
        store = GoogleCloudStorage(folder=folder)
    else:
        info("store not found")
    info(f"getting store {store}")
    return store


def get_store_from_env(folder: Optional[Path] = None) -> Optional[Store]:
    return get_store(
        type_store=TypeStore[VarEnv.TYPE_STORE.value],
        folder=folder,
    )



from .base import ParamsVarEnv
from enum import Enum

class VarEnv(ParamsVarEnv, Enum):
    PATH_TO_CRED = ParamsVarEnv(key="GCS_PATH_TO_CRED")
    NAME_BUCKET = ParamsVarEnv(key="GCS_BUCKET_NAME")
    PROJECT_BIGQUERY = ParamsVarEnv(key="GCS_BIGQUERY_PROJECT")
    DATASET_BIGQUERY = ParamsVarEnv(key="GCS_BIGQUERY_DATASET")


"""
helper functions for defining stores
"""

from pathlib import Path
from .base import Database
from .sqlite import File, MEMORY
from .google import BIGQUERY
from enum import Enum, auto
from typing import Optional
from nysmix.config.base import VarEnv
from logging import info
from sqlalchemy import case
from sqlalchemy.sql.elements import Case
from .orm import Snapshot
from nysmix.concepts import SimpleFuel


class TypeDatabase(Enum):
    MEMORY = auto()
    LOCAL = auto()
    GOOGLE = auto()


def get_database(type_database: TypeDatabase) -> Optional[Database]:
    database: Optional[Database] = None
    if type_database == TypeDatabase.MEMORY:
        database = MEMORY
    if type_database == TypeDatabase.LOCAL:
        database = File(Path("./nysmix.db"))
    elif type_database == TypeDatabase.GOOGLE:
        database = BIGQUERY
    else:
        info("database not found")
    info(f"getting database {database}")
    return database


def get_database_from_env() -> Optional[Database]:
    return get_database(
        type_database=TypeDatabase[VarEnv.TYPE_DB.value],
    )


case_simple_fuel: Case = case(
    *[simple_fuel.entry_case(Snapshot.fuel) for simple_fuel in SimpleFuel],
    else_=Snapshot.fuel,
)



from dataclasses import dataclass
from nysmix.store.base import Store
from nysmix.base import Day, from_name_file_csv
from typing import List
from logging import info
from pathlib import Path
from nysmix.database.base import Database
from sqlalchemy.future.engine import Engine
from nysmix.database.orm import mapper_registry, SnapshotStage, Snapshot
from sqlalchemy.orm import Session
from sqlalchemy import delete, select, cast, Date, insert
from sqlalchemy.sql.functions import max

NAME_TABLE_MAIN: str = "main"
NAME_TABLE_STAGE: str = "stage"


@dataclass
class Load:
    database: Database
    store_csv: Store
    num_days: int

    def create_tables(self) -> None:
        info("creating tables")
        mapper_registry.metadata.create_all(self.database.engine)

    @property
    def days_next(self) -> List[Day]:
        info("finding next csvs to load")
        days_next: List[Day] = []
        names_files_csv = [
            name_file
            for name_file in self.store_csv.names_files
            if Path(name_file).suffix == ".csv"
        ]
        if len(names_files_csv) == 0:
            info("no csvs found")
            return []
        days = [from_name_file_csv(name_file_csv) for name_file_csv in names_files_csv]
        with Session(self.database.engine) as session:
            num_records_snapshot = session.query(Snapshot).count()
        info(f"num records in Snapshot table: {num_records_snapshot}")
        if num_records_snapshot == 0:
            info("no records in Snapshot table, adding all csvs")
            days_next = days
        else:
            with Session(self.database.engine) as session:
                stmt_datetime_max = select(max(Snapshot.timestamp))
                res_datetime_max = session.execute(stmt_datetime_max)
                datetime_max = list(res_datetime_max)[0][0]
            day_max = Day(
                year=datetime_max.year, month=datetime_max.month, day=datetime_max.day
            )
            days_next = [day for day in days if day >= day_max]

        return days_next

    def stage_next(self) -> None:
        info(f"staging next days for {self}")
        days_next = sorted(self.days_next)[:self.num_days] # need to limit to avoid hitting quotas
        info(f"days next to add: {days_next}")
        with Session(self.database.engine) as session:
            num_rows_deleted = session.execute(delete(SnapshotStage).where(True))
        info(f"Rows deleted from stage: {num_rows_deleted}")
        for day in days_next:
            info(f"adding {day} to stage")
            day.set_store(self.store_csv)
            self.database.insert(csv=day.bytes_csv_from_store, table=SnapshotStage)

    def add_next(self) -> None:
        info(f"loading next days for {self}")
        self.create_tables()
        self.stage_next()

        with Session(self.database.engine) as session:
            num_records_snapshot_stage = session.query(SnapshotStage).count()
        info(f"num records in Snapshot stage table: {num_records_snapshot_stage}")

        with Session(self.database.engine) as session:
            num_records_snapshot = session.query(Snapshot).count()
        info(f"num records in Snapshot table: {num_records_snapshot}")
        if num_records_snapshot == 0:
            info("no records in Snapshot table, adding all of stage")
            stmt_select = select(
                [
                    SnapshotStage.timestamp,
                    SnapshotStage.timezone,
                    SnapshotStage.fuel,
                    SnapshotStage.gen_mw,
                ]
            )
        else:
            stmt_select = select(
                [
                    SnapshotStage.timestamp,
                    SnapshotStage.timezone,
                    SnapshotStage.fuel,
                    SnapshotStage.gen_mw,
                ]
            ).where(SnapshotStage.timestamp > select(max(Snapshot.timestamp)))

        with Session(self.database.engine) as session:
            stmt_insert = insert(Snapshot).from_select(
                [
                    Snapshot.timestamp,
                    Snapshot.timezone,
                    Snapshot.fuel,
                    Snapshot.gen_mw,
                ],
                stmt_select,
            )
            session.execute(stmt_insert)
            session.commit()

        with Session(self.database.engine) as session:
            num_records_snapshot = session.query(Snapshot).count()
        info(f"new num records in Snapshot table: {num_records_snapshot}")



"""base database class"""

from dataclasses import dataclass
from sqlalchemy.future.engine import Engine


@dataclass
class Database:

    @property
    def engine(self) -> Engine:
        pass
    
    def insert(self, csv: bytes, table: type) -> None:
        pass


# tools for working with data stores

from dataclasses import dataclass
from typing import Dict, Set
from pathlib import Path
from logging import info

@dataclass
class Store:

    @property
    def names_files(self) -> Set[str]:
        pass

    def put(self, content: bytes, name: str) -> None:
        pass

    def fetch(self, name: str) -> bytes:
        pass

    def clear(self) -> None:
        pass


@dataclass
class Local(Store):
    folder: Path

    @property
    def names_files(self) -> Set[str]:
        return set([path.name for path in self.folder.iterdir()])

    def put(self, content: bytes, name: str) -> None:
        info(f"put in store {self} under name {name}")
        with open(self.folder / name, "wb") as f:
            f.write(content)

    def fetch(self, name: str) -> bytes:
        info(f"fetch from store {self} under name {name}")
        with open(self.folder / name, "rb") as f:
            content = f.read()
        return content

@dataclass
class Memory(Store):

    def __post_init__(self):
        self.files: Dict[str, bytes] = {}

    @property
    def names_files(self) -> Set[str]:
        return set(self.files.keys())

    def put(self, content: bytes, name: str) -> None:
        info(f"put in store {self} under name {name}")
        self.files[name] =  content

    def fetch(self, name: str) -> bytes:
        info(f"fetch from store {self} under name {name}")
        return self.files[name]

    def clear(self) -> None:
        del self.files
        self.files = {}

memory = Memory()



from .base import Database
from dataclasses import dataclass
from pandas import read_csv, to_datetime
from functools import cached_property
from sqlalchemy import create_engine
from sqlalchemy.future.engine import Engine
from io import BytesIO
from pathlib import Path

@dataclass
class Sqlite(Database):

    def insert(self, csv: bytes, table: type) -> None:
        with BytesIO(csv) as f:
            df = read_csv(f)
        df.columns = ["timestamp", "timezone", "fuel", "gen_mw"]
        df["timestamp"] = to_datetime(df["timestamp"])
        df.to_sql(table.__tablename__, self.engine, if_exists="append", index=False)

@dataclass
class File(Sqlite):
    """sqlite as file"""
    path: Path

    @property
    def engine(self) -> Engine:
        return create_engine(f"sqlite:///{self.path}", echo=True, future=True)


@dataclass
class Memory(Sqlite):
    """sqlite in memory"""

    @cached_property
    def engine(self) -> Engine:
        return create_engine("sqlite://", echo=True, future=True)

MEMORY = Memory()



"""
general concepts
"""

from enum import Enum
from typing import NamedTuple, Set, Tuple
from sqlalchemy.orm.attributes import InstrumentedAttribute


class ParamsSimpleFuel(NamedTuple):
    label: str
    values: Set[str]

    def entry_case(self, field: InstrumentedAttribute) -> Tuple:
        return (field.in_(self.values), self.label)

class SimpleFuel(ParamsSimpleFuel, Enum):
    FOSSIL_FUEL = ParamsSimpleFuel(
        label="Fossil fuel", values={"Natural Gas", "Dual Fuel", "Other Fossil Fuels"}
    )
    RENEWABLE = ParamsSimpleFuel(
        label="Renewable", values={"Hydro", "Wind", "Other Renewables"}
    )
    NUCLEAR = ParamsSimpleFuel(label="Nuclear", values={"Nuclear"})



from functools import cached_property
from .base import Database
from dataclasses import dataclass
from sqlalchemy.future.engine import Engine, create_engine
from nysmix.config.google import VarEnv
from google.cloud.bigquery import Client, LoadJobConfig, WriteDisposition, SourceFormat
from google.oauth2.service_account import Credentials
from io import BytesIO
from pandas import read_csv, to_datetime


@dataclass
class BigQuery(Database):
    @property
    def engine(self) -> Engine:
        return create_engine(
            f"bigquery://{VarEnv.PROJECT_BIGQUERY.value}/{VarEnv.DATASET_BIGQUERY.value}",
            credentials_path=VarEnv.PATH_TO_CRED.value,
            echo=True,
        )

    @cached_property
    def cred(self) -> Credentials:
        return Credentials.from_service_account_file(VarEnv.PATH_TO_CRED.value)

    @cached_property
    def client(self) -> Client:
        return Client(credentials=self.cred)

    @property
    def loadjobconfig_insert(self) -> LoadJobConfig:
        return LoadJobConfig(
            write_disposition=WriteDisposition.WRITE_APPEND,
            source_format=SourceFormat.CSV,
            skip_leading_rows=1,
        )

    def insert(self, csv: bytes, table: type) -> None:
        with BytesIO(csv) as f:
            df = read_csv(f)
        df.columns = ["timestamp", "timezone", "fuel", "gen_mw"]
        df["timestamp"] = to_datetime(df["timestamp"])
        with BytesIO() as f:
            df.to_csv(f, index=False)
            f.seek(0)
            self.client.load_table_from_file(
                f,
                f"{VarEnv.PROJECT_BIGQUERY.value}.{VarEnv.DATASET_BIGQUERY.value}.{table.__tablename__}",
                job_config=self.loadjobconfig_insert,
            ).result()


BIGQUERY = BigQuery()



"""
database orm
"""

from sqlalchemy import Column, DateTime, Integer, String, ForeignKey, Float, Date, case
from sqlalchemy.orm import registry, relationship
from sqlalchemy.sql.elements import Case
from nysmix.concepts import SimpleFuel

mapper_registry = registry()

# @mapper_registry.mapped
# class Fuel:
#     __tablename__ = "fuel"
#
#     id = Column(Integer, primary_key=True)
#     name = Column(String(20))


@mapper_registry.mapped
class SnapshotStage:
    __tablename__ = "snapshot_stage"
    __schema__ = "nysmix"

    # id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, primary_key=True)
    timezone = Column(String(3), primary_key=True)
    # fuel_id = Column(Integer, ForeignKey("fuel.id")) # following https://docs.sqlalchemy.org/en/20/orm/basic_relationships.html#many-to-one
    fuel = Column(String(40), primary_key=True)  # relationship("Fuel")
    gen_mw = Column(Float)


@mapper_registry.mapped
class Snapshot:
    __tablename__ = "snapshot"
    __schema__ = "nysmix"

    # id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, primary_key=True)
    timezone = Column(String(3), primary_key=True)
    # fuel_id = Column(Integer, ForeignKey("fuel.id")) # following https://docs.sqlalchemy.org/en/20/orm/basic_relationships.html#many-to-one
    fuel = Column(String(40), primary_key=True)  # relationship("Fuel")
    gen_mw = Column(Float)


# tools for working with google cloud storage as a data store

from dataclasses import dataclass
from functools import cached_property
from .base import Store
from typing import List, Set
from logging import info
from google.oauth2.service_account import Credentials
from google.cloud.storage import Client, Bucket, Blob
from pathlib import Path
from nysmix.config.google import VarEnv


@dataclass
class GoogleCloudStorage(Store):

    folder: Path

    @cached_property
    def cred(self) -> Credentials:
        return Credentials.from_service_account_file(VarEnv.PATH_TO_CRED.value)

    @cached_property
    def client(self) -> Client:
        return Client(credentials=self.cred)

    @cached_property
    def bucket(self) -> Bucket:
        return self.client.get_bucket(VarEnv.NAME_BUCKET.value)

    @property
    def blobs(self) -> List[Blob]:
        return list(self.client.list_blobs(self.bucket, prefix=self.folder))

    @property
    def names_files(self) -> Set[str]:
        return set([Path(blob.name).name for blob in self.blobs])

    def blob(self, name: str) -> Blob:
        return self.bucket.blob(str(self.folder / name))

    def put(self, content: bytes, name: str) -> None:
        info(f"put in store {self} under name {name}")
        self.blob(name).upload_from_string(data=content)

    def fetch(self, name: str) -> bytes:
        info(f"fetch from store {self} under name {name}")
        return self.blob(name).download_as_bytes()

    def clear(self) -> None:
        info(f"clearing the bucket {self}")
        for blob in self.blobs:
            info(f"deleting blob {blob.name}")
            blob.delete()
